#include <avr/io.h>

#define F_CPU 16000000

#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "u8g.h"

// game defines
#define LCD_WIDTH  128
#define LCD_HEIGHT 64

#define UP    0
#define DOWN  1
#define LEFT  2
#define RIGHT 3
#define NONE  5

#define MAX_SIZE     300

#define BLOCK_SIZE 2

#define HEAD 1
#define BODY 2
#define TAIL 3
#define FOOD 4

#define lvl_2_x_m 35
#define lvl_2_x_M 55
#define lvl_2_y_m 12
#define lvl_2_y_M 24
#define lvl_2_x2_m 35
#define lvl_2_x2_M 80
#define lvl_2_y2_m 24
#define lvl_2_y2_M 43

#define lvl_3_x_m 10
#define lvl_3_x_M 30
#define lvl_3_y_m 8
#define lvl_3_y_M 15
#define lvl_3_x2_m 30
#define lvl_3_x2_M 37
#define lvl_3_y2_m 8
#define lvl_3_y2_M 50
#define lvl_3_x3_m 37
#define lvl_3_x3_M 73
#define lvl_3_y3_m 43
#define lvl_3_y3_M 50
#define lvl_3_x4_m 73
#define lvl_3_x4_M 80
#define lvl_3_y4_m 50
#define lvl_3_y4_M 8

typedef struct {
    uint8_t x;
    uint8_t y;
} coord;

// directia de miscare (initial spre dreapta)
uint8_t snake_direction;
// matricea ecranului
coord snake[MAX_SIZE];
//
coord food;
// pozitia in vector a capului
uint32_t head_pos;
// pozitia in vector a capului
uint32_t tail_pos;
// dimensiunea sarpelui
uint32_t snake_size;
int snake_speed;
int level;
int level_changed;
int INITIAL_SIZE;
int special_food;

// LCD
u8g_t u8g;


/* Setare directii porturi folosite */
void init_buttons(void) {

    DDRB = 0xff;
    PORTB = 0x00;

    // UP
    DDRB &= ~(1 << PB0);
    PORTB |= (1 << PB0);
    
    // LEFT
    DDRB &= ~(1 << PB1);
    PORTB |= (1 << PB1);
    
    // RIGHT
    DDRB &= ~(1 << PB2);
    PORTB |= (1 << PB2);
    
    // DOWN
    DDRB &= ~(1 << PB3);
    PORTB |= (1 << PB3);
    
    // LED user
    DDRD |= (1 << PD7);
}

/* Verificare butoane apasate */
void check_input(void) {

    // UP
    if (!(PINB & (1 << PB0)))
		if(snake_direction != UP)
			snake_direction = DOWN;
    
    // RIGHT
    if (!(PINB & (1 << PB1))) {
		if(snake_direction != LEFT)
			snake_direction = RIGHT;
    }
    
    // DOWN
    if (!(PINB & (1 << PB2))) {
		if(snake_direction != DOWN)
			snake_direction = UP;
    }
    
    // LEFT
    if (!(PINB & (1 << PB3))) {
		if(snake_direction != RIGHT)
			snake_direction = LEFT;
    }
}

void init_lcd(void) {
    MCUCR |= (1 << JTD); 
    MCUCR |= (1 << JTD);
    
    // init LCD - data ports PD0 - PD7, control on PC4 - PC1
    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
            PN(0, 0), PN(0, 1), PN(0, 2), PN(0, 3), 
            PN(0, 4), PN(0, 5), PN(0, 6), PN(0, 7), // DATA PORTS
            PN(2, 2),    // ENABLE
            U8G_PIN_NONE,
            U8G_PIN_NONE,
            PN(2, 0),    // RS / DI
            PN(2, 1),    // RW
            PN(2, 3));   // RESET
}

coord generate_coord(void) {
    coord c;
    
	c.x = rand() % 96 + 2;
	c.y = rand() % (LCD_HEIGHT - 4) + 2;
	if(level == 1) {
		return c;
	} else if(level == 2) {
		if((c.x >= lvl_2_x_m && c.x <= lvl_2_x_M && c.y >= lvl_2_y_m && c.y <= lvl_2_y_M) ||
		(c.x >= lvl_2_x2_m && c.x <= lvl_2_x2_M && c.y >= lvl_2_y2_m && c.y <= lvl_2_y2_M))
			return generate_coord();
	} else if(level == 3) {
		if((c.x >= lvl_3_x_m && c.x <= lvl_3_x_M && c.y >= lvl_3_y_m && c.y <= lvl_3_y_M) ||
		(c.x >= lvl_3_x2_m && c.x <= lvl_3_x2_M && c.y >= lvl_3_y2_m && c.y <= lvl_3_y2_M) ||
		(c.x >= lvl_3_x3_m && c.x <= lvl_3_x3_M && c.y >= lvl_3_y3_m && c.y <= lvl_3_y3_M) ||
		(c.x >= lvl_3_x4_m && c.x <= lvl_3_x4_M && c.y >= lvl_3_y4_m && c.y <= lvl_3_y4_M))
			return generate_coord();
	}
    return c;
}

void init_snake(uint32_t seed) {
    uint8_t i;
    
	snake_speed = 1;
	
    snake_size = INITIAL_SIZE;
    
    for (i = 0; i < INITIAL_SIZE; ++i) {
        snake[i].x = 6 + i * BLOCK_SIZE;
        snake[i].y = LCD_HEIGHT - LCD_HEIGHT / 16 - 8;
    }

    head_pos = INITIAL_SIZE - 1;
    tail_pos = 0;


    srand(seed);
    
    // generate food
    food = generate_coord();
}

void update_game(void) {
    uint32_t prev;

    // update head
    prev = head_pos;
    head_pos = head_pos + 1;
    
    snake[head_pos] = snake[prev];
    
    switch(snake_direction) {
        case UP:
            snake[head_pos].y = (snake[head_pos].y + BLOCK_SIZE) % LCD_HEIGHT;
            break;
        case DOWN:
            snake[head_pos].y = (snake[head_pos].y - BLOCK_SIZE) % LCD_HEIGHT;
            break;
        case LEFT:
            snake[head_pos].x = (snake[head_pos].x - BLOCK_SIZE) % LCD_WIDTH;
            break;
        case RIGHT:
            snake[head_pos].x = (snake[head_pos].x + BLOCK_SIZE) % LCD_WIDTH;
            break;
    }
    
    // check food
    if (abs(food.x - snake[head_pos].x) < BLOCK_SIZE && abs(food.y - snake[head_pos].y) < BLOCK_SIZE) {
        snake_size++;
		special_food++;
		++snake_speed;
		if(snake_speed >= 5) {
			level++;
			level_changed = 1;
		}
        food = generate_coord();
    } else {
        // update tail
        tail_pos = tail_pos + 1;
    }
}

void draw_start(void) {
    u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_helvB10);
	if(level == 1)
		u8g_DrawStr(&u8g, 20, 25, "Start eating!");
	else if(level == 2)
		u8g_DrawStr(&u8g, 20, 25, "Level 2");
	else
		u8g_DrawStr(&u8g, 20, 25, "Level 3");
}

void draw_game(void) {
    uint32_t i;
    coord block;
    int aux = BLOCK_SIZE / 2;
    u8g_SetDefaultForegroundColor(&u8g);
    for (i = 0; i < snake_size; ++i) {
        block = snake[(tail_pos + i) % MAX_SIZE];
        u8g_DrawBox(&u8g, block.x - aux, block.y - aux, BLOCK_SIZE, BLOCK_SIZE);
    }
    
    // draw food
    u8g_DrawBox(&u8g, food.x - aux, food.y - aux, BLOCK_SIZE, BLOCK_SIZE);
	
	// draw border
	for(i = 0; i < LCD_HEIGHT; i++) {
		u8g_DrawBox(&u8g, 100, i, BLOCK_SIZE, BLOCK_SIZE);
		u8g_DrawBox(&u8g, 0, i, BLOCK_SIZE, BLOCK_SIZE);
	}
	for(i = 0; i < 100; i++) {
		u8g_DrawBox(&u8g, i, LCD_HEIGHT-BLOCK_SIZE, BLOCK_SIZE, BLOCK_SIZE);
		u8g_DrawBox(&u8g, i, 0, BLOCK_SIZE, BLOCK_SIZE);
	}
	
	//draw info
	char c[4];
	sprintf(c, "s:%d", snake_speed); 
	u8g_DrawStr(&u8g, 103, 13, c);
	sprintf(c, "L:%d", level); 
	u8g_DrawStr(&u8g, 103, 28, c);
	sprintf(c, "%d", (int)snake_size); 
	u8g_DrawStr(&u8g, 103, 42, c);
	
	
	// draw obstacles
	if(level == 2) {
		int j;
		for(i = lvl_2_x_m; i < lvl_2_x_M; i++) {
			for(j = lvl_2_y_m; j < lvl_2_y_M; j++) {
				u8g_DrawBox(&u8g, i, j, 1, 1);
			}
		}
		for(i = lvl_2_x2_m; i < lvl_2_x2_M; i++) {
			for(j = lvl_2_y2_m; j < lvl_2_y2_M; j++) {
				u8g_DrawBox(&u8g, i, j, 1, 1);
			}
		}
	} else if(level == 3) {
		int j;
		for(i = lvl_3_x_m; i < lvl_3_x_M; i++) {
			for(j = lvl_3_y_m; j < lvl_3_y_M; j++) {
				u8g_DrawBox(&u8g, i, j, 1, 1);
			}
		}
		for(i = lvl_3_x2_m; i < lvl_3_x2_M; i++) {
			for(j = lvl_3_y2_m; j < lvl_3_y2_M; j++) {
				u8g_DrawBox(&u8g, i, j, 1, 1);
			}
		}
		for(i = lvl_3_x3_m; i < lvl_3_x3_M; i++) {
			for(j = lvl_3_y3_m; j < lvl_3_y3_M; j++) {
				u8g_DrawBox(&u8g, i, j, 1, 1);
			}
		}
		for(i = lvl_3_x4_m; i < lvl_3_x4_M; i++) {
			for(j = lvl_3_y4_m; j < lvl_3_y4_M; j++) {
				u8g_DrawBox(&u8g, i, j, 1, 1);
			}
		}
	}
}

void game_over(void) {
	PORTD |= (1 << PD7);
	while(1) {
		u8g_FirstPage(&u8g);
		do {
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 20, 25, "You lost!");
		} while (u8g_NextPage(&u8g));
		_delay_ms(100);
	}
}

void end_game(void) {
	PORTD |= (1 << PD7);
	while(1) {
		u8g_FirstPage(&u8g);
		do {
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 20, 25, "You win!");
		} while (u8g_NextPage(&u8g));
		_delay_ms(100);
	}
}

void check_collision(void) {
	int i;
	if(level == 1) {
		for(i = 0; i < snake_size - 1; i++) {
			if((snake[head_pos].x == snake[tail_pos + i].x &&
			   snake[head_pos].y == snake[tail_pos + i].y) ||
			   snake[head_pos].y == 0 || snake[head_pos].y == LCD_HEIGHT ||
			   snake[head_pos].x == 0 || snake[head_pos].x == 102){
					game_over();
			   }
		}
	} else if (level == 2) {
		
		for(i = 0; i < snake_size - 1; i++) {
			if((snake[head_pos].x == snake[tail_pos + i].x &&
			   snake[head_pos].y == snake[tail_pos + i].y) ||
			   snake[head_pos].y == 0 || snake[head_pos].y == LCD_HEIGHT ||
			   snake[head_pos].x == 0 || snake[head_pos].x == 102 ||
			   (snake[head_pos].x >= lvl_2_x_m && snake[head_pos].x <= lvl_2_x_M &&
			   snake[head_pos].y >= lvl_2_y_m && snake[head_pos].y <= lvl_2_y_M) ||
			   (snake[head_pos].x >= lvl_2_x2_m && snake[head_pos].x <= lvl_2_x2_M &&
			   snake[head_pos].y >= lvl_2_y2_m && snake[head_pos].y <= lvl_2_y2_M)){
					game_over();
			}
		}
	} else if (level == 3) {
		for(i = 0; i < snake_size - 1; i++) {
			if((snake[head_pos].x == snake[tail_pos + i].x &&
			   snake[head_pos].y == snake[tail_pos + i].y) ||
			   snake[head_pos].y == 0 || snake[head_pos].y == LCD_HEIGHT ||
			   snake[head_pos].x == 0 || snake[head_pos].x == 102 ||
			   (snake[head_pos].x >= lvl_3_x_m && snake[head_pos].x <= lvl_3_x_M &&
			   snake[head_pos].y >= lvl_3_y_m && snake[head_pos].y <= lvl_3_y_M) ||
			   (snake[head_pos].x >= lvl_3_x2_m && snake[head_pos].x <= lvl_3_x2_M &&
			   snake[head_pos].y >= lvl_3_y2_m && snake[head_pos].y <= lvl_3_y2_M) ||
			   (snake[head_pos].x >= lvl_3_x3_m && snake[head_pos].x <= lvl_3_x3_M &&
			   snake[head_pos].y >= lvl_3_y3_m && snake[head_pos].y <= lvl_3_y3_M) || 
			   (snake[head_pos].x >= lvl_3_x4_m && snake[head_pos].x <= lvl_3_x4_M &&
			   snake[head_pos].y >= lvl_3_y4_m && snake[head_pos].y <= lvl_3_y4_M)){
					game_over();
			}
		}
	}
}

void reset_snake(void) {
	int i;
	for(i = 0; i < MAX_SIZE; i++) {
		snake[i].x = snake[i].y = -1;
	}
}

int main(void) {

	INITIAL_SIZE = 5;

    init_buttons();
    init_lcd();
    
    snake_direction = NONE;
	level = 1;
    
    // asteptam pana la prima comanda pentru a porni jocul
	
    while(1) {
        check_input();
        
        // start game
        if (snake_direction != NONE && snake_direction != LEFT) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_start();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(100);
    }
    
    init_snake(13);

    while (1) {
        check_input();
        update_game();
		if(level_changed == 1) 
			break;
			
		check_collision();
        
        u8g_FirstPage(&u8g);
        do {
            draw_game();
        } while (u8g_NextPage(&u8g));    
        _delay_ms(60 - snake_speed * 10);
    }
	
	INITIAL_SIZE = 10;
	level_changed = 0;
	snake_direction = NONE;
	reset_snake();
    init_snake(1);
	//snake_size += 10;
	
	
	while(1) {
        check_input();
        
        // start game
        if (snake_direction != NONE && snake_direction != LEFT) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_start();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(100);
    }
	
	while (1) {
        check_input();
        update_game();
		if(level_changed == 1) 
			break;
			
		check_collision();
        
        u8g_FirstPage(&u8g);
        do {
            draw_game();
        } while (u8g_NextPage(&u8g));    
        _delay_ms(50 - snake_speed * 10);
    }
	
	INITIAL_SIZE = 15;
	level_changed = 0;
	snake_direction = NONE;
	reset_snake();
    init_snake(2);
	
	while(1) {
        check_input();
        
        // start game
        if (snake_direction != NONE && snake_direction != LEFT) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_start();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(100);
    }
	
	while (1) {
        check_input();
        update_game();
		if(level_changed == 1) 
			break;
			
		check_collision();
        
        u8g_FirstPage(&u8g);
        do {
            draw_game();
        } while (u8g_NextPage(&u8g));    
    }
	
	end_game();
    
    return 0;
}
